<?php

/*
 * Copyright by SoftCreatR Media.
 *
 * License: https://www.softcreatr.com/license-terms
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 *
 * The above copyright notice and this disclaimer notice shall be included in all
 * copies or substantial portions of the Software.
 */

namespace wcf\util;

use Exception;
use wcf\system\steam\SteamAPI;

use function array_flip;
use function chr;
use function decbin;
use function dechex;
use function defined;
use function explode;
use function extension_loaded;
use function hexdec;
use function implode;
use function md5;
use function preg_match;
use function preg_replace;
use function str_replace;
use function strlen;
use function strtr;
use function substr;
use function substr_replace;
use function trim;

final class SteamIdUtil
{
    /**
     * Base ID for calculations (32-bit compatible)
     *
     * @var int
     */
    const BASE_ID = 61197960265728;

    /**
     * List of replacement hex characters used in invite URLs
     *
     * @var string[]
     */
    public static $replacementDictionary = [
        '0' => 'b',
        '1' => 'c',
        '2' => 'd',
        '3' => 'f',
        '4' => 'g',
        '5' => 'h',
        '6' => 'j',
        '7' => 'k',
        '8' => 'm',
        '9' => 'n',
        'a' => 'p',
        'b' => 'q',
        'c' => 'r',
        'd' => 't',
        'e' => 'v',
        'f' => 'w',
    ];

    /**
     * Converts the given string to a SteamID2.
     *
     * @param $str
     * @param bool $normalize
     *
     * @return string
     *
     * @example STEAM_0:0:11101
     */
    public static function toSteamID2($str, $normalize = true)
    {
        if ($normalize === true) {
            $str = self::normalize($str);
        }

        if (self::isSteamID2($str, false)) {
            return $str;
        }

        if (self::isSteamID32($str, false)) {
            $parts = ArrayUtil::toIntegerArray(explode(':', $str));

            return 'STEAM_0:' . ((int)$str[strlen($str) - 2] % 2 === 0 ? '0:' : '1:') . (int)($parts[2] / 2);
        }

        if (self::isSteamID64($str, false)) {
            return 'STEAM_0:' . ((int)substr($str, -1) % 2 === 0 ? 0 : 1) . ':' . (int)((substr($str, 3) - self::BASE_ID) / 2);
        }

        return '';
    }

    /**
     * Returns true, if the given string is a valid steamID64.
     * Example: 76561197960287930
     *
     * @param $str
     * @param bool $normalize
     *
     * @return bool
     */
    public static function isSteamID64($str, $normalize = true)
    {
        if ($normalize === true) {
            $str = self::normalize($str);
        }

        return preg_match('/^765\d{14}$/', $str) === 1;
    }

    /**
     * Returns true, if the given string is a valid SteamID32.
     * Example: [U:1:22202]
     *
     * @param $str
     * @param bool $normalize
     *
     * @return bool
     */
    public static function isSteamID32($str, $normalize = true)
    {
        if ($normalize === true) {
            $str = self::normalize($str);
        }

        // let's allow strings without brackets to be valid
        $str = '[' . trim($str, '[]') . ']';

        return preg_match('/^\[[IUMGAPCgTLca]:[0-4]:[\d]{1,10}]$/', $str) === 1;
    }

    /**
     * Returns true, if the given string is a valid SteamID2.
     * Example: STEAM_0:0:11101
     *
     * @param $str
     * @param bool $normalize
     *
     * @return bool
     */
    public static function isSteamID2($str, $normalize = true)
    {
        if ($normalize === true) {
            $str = self::normalize($str);
        }

        return preg_match('/^STEAM_[0-4]:[0-1]:\d{1,10}$/', $str) === 1;
    }

    /**
     * Converts the given Steam user (invite) Url / ID to a SteamID64.
     *
     * @param $str
     *
     * @return string
     *
     * @example 76561197960287930
     */
    public static function fromUser($str)
    {
        $replacementDictionary = array_flip(self::$replacementDictionary);
        $code = preg_replace('~^https?://(?:www\.)?(?:steamcommunity\.com/user|s\.team/p)/([' . implode('', self::$replacementDictionary) . '-]+).*~i', '$1', StringUtil::trim($str));
        $code = str_replace('-', '', $code);
        $code = '765' . (hexdec(strtr($code, $replacementDictionary)) + self::BASE_ID);

        if (self::isSteamID64($code, false)) {
            return $code;
        }

        return '';
    }

    /**
     * Converts the given string to a SteamID32.
     *
     * @param $str
     * @param bool $normalize
     *
     * @return string
     *
     * @example [U:1:22202]
     */
    public static function toSteamID32($str, $normalize = true)
    {
        if ($normalize === true) {
            $str = self::normalize($str);
        }

        if (self::isSteamID32($str, false)) {
            return $str;
        }

        if (self::isSteamID2($str, false)) {
            $parts = ArrayUtil::toIntegerArray(explode(':', $str));

            return '[U:1:' . ($parts[1] + ($parts[2] * 2)) . ']';
        }

        if (self::isSteamID64($str, false)) {
            return '[U:1:' . (substr($str, 3) - self::BASE_ID) . ']';
        }

        return '';
    }

    /**
     * Returns basic profile information for the given string.
     *
     * @param $str
     * @param bool $normalize
     *
     * @return string[]
     */
    public static function getPlayerSummaries($str, $normalize = true)
    {
        $api = self::apiCall('GetPlayerSummaries', ['steamids' => self::toSteamID64($str, $normalize)]);

        if (!empty($api['response']['players']['player'][0])) {
            return $api['response']['players']['player'][0];
        }

        return [];
    }

    /**
     * Returns the display name for the given string,
     * or the steamID64, if none returned.
     *
     * @param $str
     * @param bool $normalize
     *
     * @return string
     *
     * @example Rabscuttle
     *
     * @deprecated Extract this information from self::getPlayerSummaries() instead
     */
    public static function getDisplayName($str, $normalize = true)
    {
        if ($normalize === true) {
            $str = self::normalize($str);
        }

        $api = self::apiCall('GetPlayerSummaries', ['steamids' => $str]);

        if (!empty($api['response']['players']['player'][0]['personaname'])) {
            return $api['response']['players']['player'][0]['personaname'];
        }

        return $str;
    }

    /**
     * Converts the given String to a Steam profile Url.
     *
     * @param $str
     * @param bool $normalize
     *
     * @return string
     *
     * @example http://steamcommunity.com/profiles/76561197960287930
     */
    public static function toProfileUrl($str, $normalize = true)
    {
        return 'https://steamcommunity.com/profiles/' . self::toSteamID64($str, $normalize) . '/';
    }

    /**
     * Converts the given string to a SteamID2.
     *
     * @param $str
     * @param bool $normalize
     *
     * @return string
     *
     * @example 76561197960287930
     */
    public static function toSteamID64($str, $normalize = true)
    {
        if ($normalize === true) {
            $str = self::normalize($str);
        }

        if (self::isSteamID64($str, false)) {
            return $str;
        }

        if (self::isSteamID2($str, false)) {
            $parts = ArrayUtil::toIntegerArray(explode(':', $str));

            return '765' . ($parts[1] + ($parts[2] * 2) + self::BASE_ID);
        }

        if (self::isSteamID32($str, false)) {
            $parts = ArrayUtil::toIntegerArray(explode(':', $str));

            return '765' . ($parts[2] + self::BASE_ID);
        }

        return '';
    }

    /**
     * Converts the given String to a Battleye GUID.
     * Requires gmp or 64-bit PHP.
     *
     * @param $str
     * @param bool $normalize
     *
     * @return string
     *
     * @example a357f31c8335a5263e0d816e64445b6a
     */
    public static function toBattleye($str, $normalize = true)
    {
        if (extension_loaded('gmp')) {
            $steamID64 = self::toSteamID64($str, $normalize);

            if (!self::isSteamID64($steamID64, false)) {
                return '';
            }

            $code = '';

            for ($i = 0; $i < 8; ++$i) {
                $code .= chr(gmp_strval(gmp_and($steamID64, 0xFF)));
                $steamID64 = gmp_div($steamID64, gmp_pow(2, 8));
            }

            return md5('BE' . $code);
        }

        if (strlen(decbin(~0)) === 64) {
            $steamID64 = self::toSteamID64($str, $normalize);

            if (!self::isSteamID64($steamID64, false)) {
                return '';
            }

            $code = '';

            for ($i = 0; $i < 8; ++$i) {
                $code .= chr($steamID64 & 0xFF);
                $steamID64 >>= 8;
            }

            return md5('BE' . $code);
        }

        return '';
    }

    /**
     * Converts the given String to a FiveM player identifier.
     *
     * @param $str
     * @param bool $normalize
     *
     * @return string
     *
     * @example steam:1100001000056ba
     */
    public static function toFiveM($str, $normalize = true)
    {
        $steamID64 = self::toSteamID64($str, $normalize);

        if (!self::isSteamID64($steamID64, false)) {
            return '';
        }

        return 'steam:1100001' . dechex((substr($steamID64, 3) - self::BASE_ID));
    }

    /**
     * Converts the given FiveM player identifier to a SteamID64.
     *
     * @param $str
     *
     * @return string
     *
     * @example 76561197960287930
     */
    public static function fromFiveM($str)
    {
        $code = '765' . (hexdec(str_replace('steam:1100001', '', $str)) + self::BASE_ID);

        if (self::isSteamID64($code, false)) {
            return $code;
        }

        return '';
    }

    /**
     * Converts the given String to a Steam user (invite) Url.
     *
     * @param $str
     * @param bool $normalize
     *
     * @return string
     *
     * @example https://s.team/p/hj-qp
     */
    public static function toUserUrl($str, $normalize = true)
    {
        $steamID64 = self::toSteamID64($str, $normalize);

        if (!self::isSteamID64($steamID64, false)) {
            return '';
        }

        $code = dechex((substr($steamID64, 3) - self::BASE_ID));
        $code = strtr($code, self::$replacementDictionary);
        $length = strlen($code);

        if ($length > 3) {
            $code = substr_replace($code, '-', (int)($length / 2), 0);
        }

        return 'https://s.team/p/' . $code . '/';
    }

    /**
     * Returns the universe for the given Steam ID.
     *
     * The value of X is 0 in VALVe's GoldSrc and Source Orange Box Engine games (For example, Counter Strike: Source),
     * but newer Valve games such as Left 4 Dead, Left 4 Dead 2 and Alien Swarm have 1 as a value of X.
     * The value of Y is either 0 or 1 and is part of the Account ID as described above.
     * This bit used to indicate which authentication server was used by that account.
     *
     * 0 = Individual / Unspecified / Invalid
     * 1 = Public
     * 2 = Beta
     * 3 = Internal
     * 4 = Dev
     * 5 = RC
     *
     * @param $str
     * @param bool $normalize
     *
     * @return int
     */
    public static function getUniverse($str, $normalize = true)
    {
        if ($normalize === true) {
            $str = self::normalize($str);
        }

        if (self::isSteamID2($str)) {
            $parts = ArrayUtil::toIntegerArray(explode(':', $str));

            if ($parts[1] === 0) {
                $parts[1] = 1;
            }

            return $parts[1];
        }

        return -1;
    }

    /**
     * Returns the account type for the given Steam ID.
     *
     * There are ten known account types for a Steam account, of which 4 can be persistently created today.
     *
     * Users of an 'Individual' account are temporarily referred to as having a 'Pending' account which
     * have a textual representation of "STEAM_ID_PENDING" until their account credentials are verified with
     * Steam's authentication servers, a process usually complete by the time a server is fully connected to.
     * Accounts of the type 'Invalid' have a textual representation of "UNKNOWN" and are used for bots and
     * accounts which do not belong to another class.
     *
     * Multi-user chats use the 'T' character.
     * Steam group (clan) chats use the 'c' character.
     * Steam lobbies use Chat IDs and use the 'L' character.
     *
     * I = Invalid
     * U = Individual
     * M = Multiseat
     * G = GameServer
     * A = AnonGameServer
     * P = Pending
     * C = ContentServer
     * g = Clan
     * T = Chat
     * c = Chat
     * L = Chat
     * a = AnonUser
     *
     * @param $str
     * @param bool $normalize
     *
     * @return string
     */
    public static function getAccountType($str, $normalize = true)
    {
        if ($normalize === true) {
            $str = self::normalize($str);
        }

        if (self::isSteamID32($str)) {
            $parts = ArrayUtil::toIntegerArray(explode(':', $str));

            return $parts[1];
        }

        return '';
    }

    /**
     * Returns all player bans (Community Ban, VAC Status, ...) for the given Steam ID.
     *
     * @param $str
     * @param bool $normalize
     *
     * @return array
     */
    public static function getPlayerBans($str, $normalize = true)
    {
        if ($normalize === true) {
            $str = self::normalize($str);
        }

        $api = self::apiCall('GetPlayerBans', ['steamids' => $str]);

        if (!empty($api['players'][0])) {
            unset($api['players'][0]['SteamId']); // unnecessary

            return $api['players'][0];
        }

        return [];
    }

    /**
     * Normalizes the given string.
     * If the given string is an Url, we will try to extract the SteamID64.
     *
     * For performance reasons, this method shouldn't be called more than once.
     *
     * @param $str
     *
     * @return string
     *
     * @noinspection NotOptimalRegularExpressionsInspection
     */
    private static function normalize($str)
    {
        $str = StringUtil::trim($str);

        if (empty($str) || self::isSteamID64($str, false) || self::isSteamID32($str, false) || self::isSteamID2($str, false)) {
            return $str;
        }

        if (preg_match('~^https?://(?:www\.)?(?:steamcommunity\.com/user|s\.team/p)/([' . implode('', self::$replacementDictionary) . '-]+).*~i', $str)) {
            return self::fromUser($str);
        }

        if (preg_match('~^https?://(?:www\.)?steamcommunity\.com/profiles/\d{17}.*~i', $str)) {
            return preg_replace('~^https?://(?:www\.)?steamcommunity\.com/profiles/(\d{17}).*~i', '$1', $str);
        }

        // ultimative fallback
        // if this doesn't return a SteamID64, the user just provided garbage
        if (preg_match('~(https?://(?:www\.)?steamcommunity\.com/id/)?[a-zA-Z\d_-]{2,32}.*~i', $str)) {
            $api = self::apiCall('ResolveVanityURL', [
                'vanityurl' => preg_replace('~https?://(?:www\.)?steamcommunity\.com/id/([a-zA-Z\d_-]{2,32}).*~i', '$1', $str),
                'url_type' => 1
            ]);

            if (!empty($api['response']['steamid'])) {
                return $api['response']['steamid'];
            }
        }

        return '';
    }

    /**
     * Wrapper for the Steam Web Api class by Hanashi.
     *
     * @param $method
     *
     * @return array
     */
    private static function apiCall($method, array $data = [])
    {
        if (defined('HA_STEAM_API_KEY') && HA_STEAM_API_KEY !== '') {
            try {
                return SteamAPI::execute('ISteamUser', $method, 1, $data);
            } catch (Exception $e) {
                // ignore
            }
        }

        return [];
    }

    /**
     * Forbid creation of SteamIdUtil objects.
     */
    private function __construct()
    {
        // does nothing
    }
}
